(*

; const
    maxd	= 500       
  ; maxorbits	= 1400
  ; maxparams	= 1200

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; matrix3x3	= array[1..3,1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end

; var
    d,orbits,params	: integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer

*)

(**********************************************************************)
(*                                                                    *)
(*                 basic routines involving integers                  *)
(*                                                                    *)
(**********************************************************************)

; function sgn ( a : integer ) : sign
  ; begin
      if a>0
      then sgn := 1
      else
      if a<0
      then sgn := -1
      else sgn := 0
    end (* sgn *)

; function modd ( a : integer ) : integer
  (* The smallest positive representative of a mod d. *)
  ; begin
      a := a mod d
    ; if a>0
      then modd := a
      else modd := a+d
    end (* modd *)

; procedure swapints ( var a,b : integer )
  (* Swap two integers. *)
  ; var
      h : integer
  ; begin
      h := a
    ; a := b
    ; b := h
    end (* swapints *)

; function min3 ( a,b,c : integer ) : integer
  (* The minimum of three integers. *)
  ; begin
      if a<b then swapints(a,b)
    ; if b<c then swapints(b,c)
    ; min3 := c
    end (* min3 *)

; function gcd2 ( a,b : integer ) : integer
  (* The g.c.d. of two integers. *)
  ; begin
      a := abs(a)
    ; b := abs(b)
    ; if a<b then swapints(a,b)
    ; while b<>0 do
      begin
        a := a mod b
      ; swapints(a,b)
      end
    ; gcd2 := a
    end (* gcd2 *)

; function lcm2 ( a,b : integer ) : integer
  (* The l.c.m. of two integers. *)
  ; begin
      if (a=0) or (b=0)
      then lcm2 := 0
      else lcm2 := (a*b) div gcd2(a,b)
    end (* lcm2 *)

; function gcd3 ( a,b,c : integer ) : integer
  (* The g.c.d. of three integers. *)
  ; begin
      a := abs(a)
    ; b := abs(b)
    ; c := abs(c)
    ; if a<b then swapints(a,b)
    ; if b<c then swapints(b,c)
    ; while c<>0 do
      begin
        a := a mod c
      ; b := b mod c
      ; if a<b
        then swapints(a,c)
        else swapints(b,c)
      end
    ; gcd3 := gcd2(a,b)
    end

; function coprimewithd ( a,b,c : integer ) : boolean
  (* Test whether three integers and d are coprime. *)
  ; begin
      a := modd(abs(a))
    ; b := modd(abs(b))
    ; c := modd(abs(c))
    ; coprimewithd := ( gcdwithd[gcd3(a,b,c)] = 1 )
    end (* coprimewithd *)

(**********************************************************************)
(*                                                                    *)
(*                    making some auxiliary tables                    *)
(*                                                                    *)
(**********************************************************************)

; procedure make2tables
  (* Make a table containing the g.c.d. of every integer with d, *)
  (* and a table containing "inverse integers" of every integer. *)
  ; var
      a : integer
    ; b1,s1,b2,s2,q : integer
  ; begin
      for a := 1 to d do
      begin
        b1 := 0
      ; s1 := d (* b1*a cong. s1 mod d *)
      ; b2 := 1
      ; s2 := a (* b2*a cong. s2 mod d *)
      ; repeat
          q := s1 div s2
        ; b1 := b1-q*b2
        ; s1 := s1-q*s2
        ; swapints(b1,b2)
        ; swapints(s1,s2)
        until s2=0
      ; gcdwithd[a] := s1
      ; invmodd[a] := b1
      end
    ; for a := 1 to d do
      begin
        b1 := invmodd[a] mod (d div gcdwithd[a])
      ; if b1<=0 then b1 := b1+(d div gcdwithd[a])
      ; while gcdwithd[b1]>1 do b1 := b1+(d div gcdwithd[a])
      ; invmodd[a] := b1
      end
    end (* make2tables *)

(**********************************************************************)
(*                                                                    *)
(*           basic procedures involving (signed) 3-vectors            *)
(*                                                                    *)
(**********************************************************************)

; procedure makevec ( x1,x2,x3 : integer ; var v : vector3 )
  (* Put x1,x2,x3 into a vector. *)
  ; begin
      v[1] := x1
    ; v[2] := x2
    ; v[3] := x3
    end (* makevec *)

; procedure makevecmodd ( x1,x2,x3 : integer ; var v : vector3 )
  (* Put x1,x2,x3 into a vector and reduce it modulo d. *)
  ; begin
      v[1] := modd(x1)
    ; v[2] := modd(x2)
    ; v[3] := modd(x3)
    end (* makevecmodd *)

; procedure plus ( v : vector3 ; var sv : sgdvector3 )
  (* Provide a 3-vector with a plus sign. *)
  ; begin
      sv.x := v
    ; sv.sg := 1
    end (* plus *)

; function prec ( v1,v2 : vector3 ) : boolean
  (* Test whether the first vector precedes the  *)
  (* second one in the lexicographical ordering. *)
  ; begin
      prec := false
    ; if v1[1]<v2[1]
      then prec := true
      else
        if v1[1]=v2[1] then
        if v1[2]<v2[2]
        then prec := true
        else
          if v1[2]=v2[2] then
          if v1[3]<v2[3]
          then prec := true
    end (* prec *)

; function oppo ( sv1,sv2 : sgdvector3 ) : boolean
  (* Test whether two signed 3-vectors are opposite. *)
  ; begin
      oppo := (sv1.x[1]=sv2.x[1]) and (sv1.x[2]=sv2.x[2])
        and (sv1.x[3]=sv2.x[3]) and (sv1.sg+sv2.sg=0)
    end (* oppo *)

; procedure multscalvec ( sc : integer ; sv : sgdvector3
  ; var sp : sgdvector3 )
  (* Multiply a signed 3-vector by a scalar. *)
  ; var
      i : integer
  ; begin
      for i := 1 to 3 do sp.x[i] := modd(sc*sv.x[i])
    ; sp.sg := sv.sg
    end (* multscalvec *)

; procedure movetox1 ( i : integer ; sv : sgdvector3
  ; var mv : sgdvector3 )
  (* Move x[i] to x[1]. *)
  ; begin
      if i>1 then
      with sv do
      begin
        swapints(x[i],x[1])
      ; sg := -sg
      end
    ; mv := sv
    end (* xitox1 *)

(**********************************************************************)
(*                                                                    *)
(*                 finding the minimal representative                 *)
(*                                                                    *)
(**********************************************************************)

; procedure actd8 ( sv : sgdvector3 ; var min : sgdvector3
                  ; var opp : boolean )
  (* Look at the orbit of sv under D8.  Return a   *)
  (* minimal element in min.  Return in opp        *)
  (* whether the orbit contains opposite elements. *)
  ; begin
      with sv do
      begin
        if modd(-x[2])<x[2] then x[2] := modd(-x[2])
      ; if modd(-x[3])<x[3] then x[3] := modd(-x[3])
      ; if x[3]<x[2] then
        begin
          swapints(x[3],x[2])
        ; sg := -sg
        end
      ; opp := ( x[2]=x[3] )
      end
    ; min := sv
    end (* actd8 *)

; procedure actd8xstab ( sv : sgdvector3 ; var min : sgdvector3
                         ; var opp : boolean )
  (* Look at the orbit of sv under                   *)
  (* D8 x ( stabilizer of x[1] in (Z/dZ)* ).         *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      sc : integer
    ; sp,mini : sgdvector3
    ; oppi : boolean
  ; begin
      min := sv
    ; opp := false
    ; sc := 1 (* sc will run through integers cong. 1 mod d/sv.x[1] *)
    ; repeat
        if gcdwithd[sc]=1 then (* it's a unit *)
        begin
          multscalvec(sc,sv,sp)
        ; actd8(sp,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      ; sc := sc+(d div sv.x[1])
      until sc>d
    end (* actd8xstab *)

; procedure acts4xmul ( sv : sgdvector3 ; var min : sgdvector3
                        ; var opp : boolean )
  (* Look at the orbit of sv under S4 x (Z/dZ)*.     *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      mg,i : integer
    ; sp,mv,mini : sgdvector3
    ; oppi : boolean
  ; begin
      with sv do
      begin
        mg := min3(gcdwithd[x[1]],gcdwithd[x[2]],gcdwithd[x[3]])
      ; min := sv
      ; opp := false
      ; for i := 1 to 3 do
        begin
          multscalvec(invmodd[x[i]],sv,sp)
        ; movetox1(i,sp,mv)
        ; actd8xstab(mv,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      end
    end (* acts4xmul *)

(**********************************************************************)
(*                                                                    *)
(*                     using  the list of orbits                      *)
(*                                                                    *)
(**********************************************************************)

; function orbitnum ( v : vector3 ) : integer
  (* The number in the list of orbits of a minimal representative. *)
  ; var
      lw,md,up : integer
  ; begin
      lw := 1
    ; up := orbits+1
    ; repeat
        md := (lw+up) div 2
      ; if prec(v,orbit[md].x)
        then up := md
        else lw := md
      until up=lw+1
    ; orbitnum := lw
    end (* orbitnum *)

; procedure locate ( v : vector3 ; var on,pr : integer)
  (* Find the orbit number and parameter reference. *)
  ; var
      plusv,min : sgdvector3
    ; opp : boolean
  ; begin
      plus(v,plusv)
    ; acts4xmul(plusv,min,opp)
    ; on := orbitnum(min.x)
    ; pr := min.sg*orbit[on].parref
    end (* parrefofvec *)

(**********************************************************************)
(*                                                                    *)
(*                     making a list of variables                     *)
(*                                                                    *)
(**********************************************************************)

; procedure makeparams
  (* Make a list of parameters, i.e. of orbits whose function      *)
  (* values are not known to vanish or to be equal to the function *)
  (* value of some other orbit.  Set the parref field of orbit.    *)
  ; var
      i : integer
  ; begin
      params := 0
    ; for i := 1 to orbits do
      with orbit[i] do
      if abs(orbref)<i
      then
      begin
        if orbref<>0
        then parref := sgn(orbref)*orbit[abs(orbref)].parref
      end
      else
      begin
        params := params+1
      ; parref := params
      ; param[params] := i
      end
    end (* makeparams *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure readorbits
  ; var
      olf : text
    ; olfname : varying [100] of char (* SUN Pascal *)
    ; i : integer
  ; begin
      argv(1,olfname)
    ; reset(olf,olfname) (* SUN Pascal *)
    ; readln(olf,d)
    ; readln(olf,orbits)
    ; for i := 1 to orbits do
      with orbit[i] do
      readln(olf,x[1],x[2],x[3],orbref)
    end
