program hecke ( input,output )

; const
    maxd	= 500       (* changed by Bert, 14-9-93, was 200 *)
    (* maxd	= 200 *)
  ; maxorbits	= 5000
  ; maxparams	= 5000
  ; maxdim	= 200     (* changed 4-9-93 by Bert *)
  ; tab		= '	' (* TAB *)

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; matrix3x3	= array[1..3,1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end

; var
    d,orbits,params,dim,ell	: integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer
  ; soln		: array [1..maxdim,1..maxparams] of integer
  ; freepar		: array [1..maxdim] of integer
  ; raw			: array [1..maxdim,1..maxparams] of integer
  ; enum		: array [1..maxdim,1..maxdim] of integer
  ; denomef		: integer

#include "common.p"

(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

; function det3x3 ( a : matrix3x3 ) : integer
  (* The determinant of a 3 x 3 matrix. *)
  ; begin
(*    det3x3 := + a[1,1]*a[2,2]*a[3,3] - a[1,1]*a[2,3]*a[3,2]
                - a[1,2]*a[2,1]*a[3,3] + a[1,2]*a[2,3]*a[3,1]
                + a[1,3]*a[2,1]*a[3,2] - a[1,3]*a[2,2]*a[3,1] *)
      det3x3 := + a[1,1]*(a[2,2]*a[3,3]-a[2,3]*a[3,2])
                - a[1,2]*(a[2,1]*a[3,3]-a[2,3]*a[3,1])
                + a[1,3]*(a[2,1]*a[3,2]-a[2,2]*a[3,1])
    end (* det3x3 *)

; procedure mult3x3 ( m1,m2 : matrix3x3 ; var pm : matrix3x3 )
  (* The product of two 3 x 3 matrices. *)
  ; var
      i,j,k : integer
  ; begin
      for i := 1 to 3 do
      for k := 1 to 3 do
      begin
        pm[i,k] := 0
      ; for j := 1 to 3 do pm[i,k] := pm[i,k]+m1[i,j]*m2[j,k]
      end
    end (* mult3x3 *)

; procedure multvec3x3 ( v : vector3 ; m : matrix3x3
  ; var p : vector3 )
  ; var
      i,j : integer
  ; begin
      for j := 1 to 3 do
      begin
        p[j] := 0
      ; for i := 1 to 3 do p[j] := p[j]+v[i]*m[i,j]
      end
    end (* multvec3x3 *)

; procedure swapcols3x3 ( c1,c2 : integer ; var m : matrix3x3 )
  (* Swap two columns of a 3 x 3 matrix. *)
  ; var
      i : integer
  ; begin
      for i := 1 to 3 do swapints(m[i,c1],m[i,c2])
    end (* swapcols3x3 *)

; procedure elimcols3x3 ( r,c1,c2 : integer ; var m : matrix3x3 )
  (* An elimination step on the columns of a 3 x 3 matrix. *)
  ; var
      i,q : integer
  ; begin
      q := m[r,c2] div m[r,c1]
    ; for i := 1 to 3 do m[i,c2] := m[i,c2]-q*m[i,c1]
    end (* elimcols3x3 *)

; procedure swapsim3and3x3 ( i,j : integer ; var v : vector3
    ; var m : matrix3x3 ; var detm : integer )
  (* Simultaneously swap the i-th and j-th coordinate of a
     3-vector, and the i-th and j-th column of a 3 x 3 matrix. *)
  ; var
      k : integer
  ; begin
      swapints(v[i],v[j])
    ; for k := 1 to 3 do swapints(m[k,i],m[k,j])
    ; detm := -detm
    end (* swapsim3and3x3 *)

; procedure redsim3and3x3 ( i,j : integer ; var v : vector3
    ; var m : matrix3x3 )
  (* Reduce the j-th coordinate of a 3-vector modulo
     its i-th coordinate, and apply the opposite
     operation to the columns of a 3 x 3 matrix. *)
  ; var
      k,q : integer
  ; begin
      q := v[j] div v[i]
    ; v[j] := v[j]-q*v[i]
    ; for k := 1 to 3 do m[k,i] := m[k,i]+q*m[k,j]
    end (* redsim3and3x3 *)

; procedure ext3to3x3 ( v : vector3 ; var xm : matrix3x3 )
  (* Find a unimodular matrix whose first column equals v.
     The algorithm is due to F. Beukers.                            *)
  ; var
      i,j,detxm : integer
  ; begin
      for i := 1 to 3 do
      for j := 1 to 3 do
      xm[i,j] := 0
    ; for i := 1 to 3 do
      xm[i,i] := 1
    ; detxm := 1
    ; if abs(v[1])<abs(v[2]) then swapsim3and3x3(1,2,v,xm,detxm)
    ; if abs(v[2])<abs(v[3]) then swapsim3and3x3(2,3,v,xm,detxm)
    ; while v[3]<>0 do
      begin
        for i := 1 to 2 do redsim3and3x3(3,i,v,xm)
      ; if abs(v[1])<abs(v[2])
        then swapsim3and3x3(1,3,v,xm,detxm)
        else swapsim3and3x3(2,3,v,xm,detxm)
      end
    ; if abs(v[1])<abs(v[2]) then swapsim3and3x3(1,2,v,xm,detxm)
    ; while v[2]<>0 do
      begin
        redsim3and3x3(2,1,v,xm)
      ; swapsim3and3x3(1,2,v,xm,detxm)
      end
    ; if v[1]=-1
      then
      begin
        for i := 1 to 3 do xm[i,1] := -xm[i,1]
      ; detxm := -detxm
      end
    ; if detxm=-1
      then for i := 1 to 3 do swapints(xm[i,2],xm[i,3])
    end (* ext3to3x3 *)

(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

; function smallrep ( a,m : integer ) : integer
  (* A small representative of a mod m. *)
  ; begin
      m := abs(m)
    ; a := a mod m
    ; if a>m div 2
      then smallrep := a-m
      else 
      if a<-m div 2
      then smallrep := a+m
      else smallrep := a
    end (* smallrep *)

; procedure goodvec ( m : matrix3x3 ; var gv : vector3 )
  (* Find a suitable vector for the reduction of a modular
     symbol to a sum of symbols with smaller determinant. *)
  ; var
      mred : matrix3x3
    ; x : vector3
    ; detm,denom,i : integer
  ; begin
      detm := det3x3(m)
    ; mred := m
      (* Perform Euclid--Gauss elimination on the first row. *)
    ; if abs(mred[1,1])<abs(mred[1,2]) then swapcols3x3(1,2,mred)
    ; if abs(mred[1,2])<abs(mred[1,3]) then swapcols3x3(2,3,mred)
    ; while mred[1,3]<>0 do
      begin
        elimcols3x3(1,3,1,mred)
      ; elimcols3x3(1,3,2,mred)
      ; if abs(mred[1,1])<abs(mred[1,2])
        then swapcols3x3(1,3,mred)
        else swapcols3x3(2,3,mred)
      end
    ; if abs(mred[1,1])<abs(mred[1,2]) then swapcols3x3(1,2,mred)
    ; while mred[1,2]<>0 do
      begin
        elimcols3x3(1,2,1,mred)
      ; swapcols3x3(1,2,mred)
      end
      (* The Euclid--Gauss elimination on the first row is done. *)
    ; if abs(mred[1,1])>1
      then
      begin
        x[3] := 0
      ; x[2] := 0
      ; x[1] := 1
      ; denom := mred[1,1]
      end
      else
      begin
        (* Perform Euclid--Gauss elimination on the second row. *)
        if abs(mred[2,2])<abs(mred[2,3]) then swapcols3x3(2,3,mred)
      ; while mred[2,3]<>0 do
        begin
          elimcols3x3(2,3,2,mred)
        ; swapcols3x3(2,3,mred)
        end
        (* The Euclid--Gauss elimination on the second row is done. *)
      ; if abs(mred[2,2])>1
        then
        begin
          x[3] := 0
        ; x[2] := 1
        ; x[1] := -mred[1,1]*mred[2,1] (* *x[2] *)
        ; denom := mred[2,2]
        end
        else (* We have |mred[3,3]| = |det(m)| > 1. *)
        begin
          x[3] := 1
        ; x[2] := -mred[2,2]*mred[3,2] (* *x[3] *)
        ; x[1] := -mred[1,1]*(mred[2,1]*x[2]+mred[3,1](* *x[2] *))
        ; denom := mred[3,3]
        end
      end
      (* We have found a vector x such that x.m cong. 0 mod denom. *)
    ; for i := 1 to 3 do x[i] := smallrep(x[i],denom)
    ; multvec3x3(x,m,gv)
    ; for i := 1 to 3 do gv[i] := gv[i] div denom
    end (* goodvec *)

; procedure reprow ( r : integer ; m : matrix3x3 ; v : vector3
                  ; var mv : matrix3x3 )
  (* Replace a row of a 3 x 3 matrix by another vector. *)
  ; var
      i : integer
  ; begin
      mv := m
    ; for i := 1 to 3 do mv[r,i] := v[i]
    end (* reprow *)

; procedure addunimodtorow ( r : integer ; ms : matrix3x3 )
  (* Add a unimodular symbol to a row of the raw Hecke matrix. *)
  ; var
      i,on,pr : integer
    ; v : vector3
  ; begin
      for i := 1 to 3 do v[i] := modd(ms[i,1])
    ; locate(v,on,pr)
    ; if pr<>0 then raw[r,abs(pr)] := raw[r,abs(pr)]+sgn(pr)
    end (* addunimodtorow *)

; procedure addmodsymtorow ( ms : matrix3x3 ; r : integer )
  (* Add a modular symbol to a row of the raw Hecke
     matrix.  In order to do this, reduce the modular
     symbol to a sum of unimodular symbols. *)
  ; var
      i,j,g,detms : integer
    ; mv : matrix3x3
    ; v : vector3
  ; begin
      if det3x3(ms)<>0
      then
      begin
        for i := 1 to 3 do
        begin
          g := gcd3(ms[i,1],ms[i,2],ms[i,3])
        ; if g>1 then for j := 1 to 3 do ms[i,j] := ms[i,j] div g
        end
        (* Now all rows of ms are primitive. *)
      ; detms := det3x3(ms)
      ; if abs(detms)=1
        then
        begin
          if detms=-1
          then for i := 1 to 3 do ms[1,i] := -ms[1,i]
        ; addunimodtorow(r,ms)
        end
        else
        begin
          goodvec(ms,v)
        ; for i := 1 to 3 do
          begin
            reprow(i,ms,v,mv)
          ; addmodsymtorow(mv,r)
          end
        end
      end
    end (* addmodsymtorow *)

; procedure make3x3 ( m11,m12,m13,m21,m22,m23,m31,m32,m33 : integer
                    ; var m : matrix3x3 )
  ; begin
      m[1,1] := m11
    ; m[1,2] := m12
    ; m[1,3] := m13
    ; m[2,1] := m21
    ; m[2,2] := m22
    ; m[2,3] := m23
    ; m[3,1] := m31
    ; m[3,2] := m32
    ; m[3,3] := m33
    end (* thematrix3x3 *)
